import numpy as np
import torch
from torch.utils.data import Dataset


def generate_iid_data(
    size, shape, loc=0, scale=1, local_signal=0, local_size=None, seed=0
):
    """Generate synthetic data with iid Gaussian noise.

    Parameters
    ----------
    size : int
        The number of samples.
    shape : int or tuple
        The shape of the data.
    loc : float, optional
        The mean of the Gaussian noise. Default is 0.0.
    scale : float, optional
        The standard deviation of the Gaussian noise. Default is 1.0.
    signal : float, optional
        The signal strength. Default is 1.0.
    seed : int, optional
        The seed of the random number generator. Default is None.

    Returns
    -------
    np.ndarray
        The generated data.
    np.ndarray
        The masks of the signal.
    np.ndarray
        The labels indicating the presence of the signal.
    """
    if local_size is None:
        local_size = shape[-1] // 3

    rng = np.random.default_rng(seed=seed)

    if isinstance(shape, int):
        shape = (shape,)

    data = rng.normal(loc, scale, (size, *shape))

    masks = np.zeros((size, *shape))

    for dim in shape[1:]:
        assert (
            local_size < dim
        ), "local_size must be less than to the shape of the data."

    local_shape = tuple(max(dim, local_size) for dim in shape[1:])

    local_positions = [
        tuple(rng.integers(0, max(1, data_dim - local_dim), size))
        for data_dim, local_dim in zip(shape, local_shape)
    ]

    for i in range(size):
        slices = tuple(
            slice(pos[i], pos[i] + local_dim)
            for pos, local_dim in zip(local_positions, local_shape)
        )
        data[i][slices] += local_signal
        if local_signal != 0:
            masks[i][slices] = 1

    labels = (local_signal != 0) * np.ones(size, dtype=int)

    return data, masks, labels


class SyntheticDataset(Dataset):
    """A PyTorch Dataset for generating synthetic data with iid Gaussian noise and local signals.

    This dataset generates samples containing Gaussian noise with optional local signals
    added at random positions.

    The dataset generates three components for each sample:
        1. Data: Tensor containing Gaussian noise with optional local signals
        2. Mask: Binary tensor indicating the position of local signals
        3. Label: Binary value indicating presence of local signal

    Attributes
    ----------
    data : numpy.ndarray
        Array containing the generated synthetic data.
    masks : numpy.ndarray
        Array containing the masks for local signals.
    labels : numpy.ndarray
        Array containing the binary labels.

    Examples
    --------
    >>> dataset = SyntheticDataset(
    ...     size=1000,
    ...     shape=(32, 32),
    ...     loc=0,
    ...     scale=1,
    ...     local_signal=2.0
    ... )
    >>> data, mask, label = dataset[0]
    >>> print(data.shape, mask.shape, label.shape)
    torch.Size([32, 32]) torch.Size([32, 32]) torch.Size([])
    """

    def __init__(
        self,
        size,
        shape,
        loc=0,
        scale=1,
        local_signal=0,
        local_size=None,
        seed=0,
    ):
        self.data, self.masks, self.labels = generate_iid_data(
            size,
            shape,
            loc,
            scale,
            local_signal,
            local_size,
            seed,
        )

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data = torch.from_numpy(self.data[idx]).float()
        mask = torch.from_numpy(self.masks[idx]).float()
        label = torch.tensor(self.labels[idx]).float()
        return data, mask, label



import torch
import torch.nn as nn
import torch.nn.functional as F
from .utils import truncated_interval


class InputDiff:
    def __init__(self, *args, **kwargs):
        """
        The class computes the difference with the input that is selected 'o_idx'.
        """
        super().__init__()

    def forward(self, x, input_x):
        return x - input_x

    def forward_si(self, x, a, b, l, u, z, input_x, input_a, input_b):
        """Compute (x - input_x), (a - input_a), (b - input_b)

        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            input tensor or tensor list
        a : torch.Tensor | list[torch.Tensor]
            a tensor or tensor list
        b : torch.Tensor | list[torch.Tensor]
            b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            l tensor or tensor list
        u : torch.Tensor | list[torch.Tensor]
            u tensor or tensor list
        z : float
        Input_x : torch.Tensor | list[torch.Tensor]
            input tensor or tensor list
        Input_a : torch.Tensor | list[torch.Tensor]
            a tensor or tensor list
        Input_b : torch.Tensor | list[torch.Tensor]
            b tensor or tensor list

        Returns
        -------
        output_x : torch.Tensor | list[torch.Tensor]
            output tensor or tensor list
        output_a : torch.Tensor | list[torch.Tensor]
            output a tensor or tensor list
        output_b : torch.Tensor | list[torch.Tensor]
            output b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            lower bound of the truncated interval
        u : torch.Tensor | list[torch.Tensor]
            upper bound of the truncated interval
        """
        x = x - input_x
        a = a - input_a
        b = b - input_b
        return x, a, b, l, u


class Abs:
    def __init__(self, *args, **kwargs):
        """The class computes the absolute value of the input and the interval of the truncated interval."""
        super().__init__()

    def forward(self, x):
        return torch.abs(x)

    def forward_si(self, x, a, b, l, u, z):
        """Compute the interval [l, u] = {|a_i + b_i * z| > 0}

        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            input tensor or tensor list
        a : torch.Tensor | list[torch.Tensor]
            a tensor or tensor list
        b : torch.Tensor | list[torch.Tensor]
            b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            l tensor or tensor list
        u : torch.Tensor | list[torch.Tensor]
            u tensor or tensor list
        z : float

        Returns
        -------
        output_x : torch.Tensor | list[torch.Tensor]
            output tensor or tensor list
        output_a : torch.Tensor | list[torch.Tensor]
            output a tensor or tensor list
        output_b : torch.Tensor | list[torch.Tensor]
            output b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            lower bound of the truncated interval
        u : torch.Tensor | list[torch.Tensor]
            upper bound of the truncated interval
        """

        # Compute the interval |a_i + b_i * z| > 0 -> a_i + b_i * z < 0
        negative_index = x < 0
        tTa = a
        tTb = b
        tTa = torch.where(negative_index, -tTa, tTa)
        tTb = torch.where(negative_index, -tTb, tTb)
        l_negative, u_negative = truncated_interval(tTa, tTb)
        l = torch.max(l, l_negative)
        u = torch.min(u, u_negative)
        assert l <= z <= u

        # Compute the interval |a_i + b_i * z| > 0 -> a_i + b_i * z_i > 0
        positive_index = x > 0
        tTa = a
        tTb = b
        tTa = torch.where(positive_index, tTa, -tTa)
        tTb = torch.where(positive_index, tTb, -tTb)
        l_positive, u_positive = truncated_interval(tTa, tTb)
        l = torch.max(l, l_positive)
        u = torch.min(u, u_positive)
        assert l <= z <= u

        output_x = torch.abs(x)
        output_a = torch.where(positive_index, a, -a)
        output_b = torch.where(positive_index, b, -b)
        return output_x, output_a, output_b, l, u


class AverageFilter:
    def __init__(self, kernel_size: int = 3):
        super().__init__()
        self.kernel_size = kernel_size
        self.padding = kernel_size // 2

    def _create_kernel(self, dim):
        if dim == 3:  # 1D input
            kernel = torch.ones((1, 1, self.kernel_size), dtype=torch.float64)
            kernel = kernel / self.kernel_size
        else:  # 2D input
            kernel = torch.ones(
                (1, 1, self.kernel_size, self.kernel_size), dtype=torch.float64
            )
            kernel = kernel / (self.kernel_size**2)
        return nn.Parameter(kernel, requires_grad=False)

    def forward(self, x):
        kernel = self._create_kernel(x.dim())
        conv_func = {3: F.conv1d, 4: F.conv2d}[x.dim()]
        return conv_func(x, kernel, padding=self.padding, groups=x.shape[1])

    def forward_si(self, x, a, b, l, u, z):
        """
        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            input tensor or tensor list
        a : torch.Tensor | list[torch.Tensor]
            a tensor or tensor list
        b : torch.Tensor | list[torch.Tensor]
            b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            l tensor or tensor list
        u : torch.Tensor | list[torch.Tensor]
            u tensor or tensor list
        z : float

        Returns
        -------
        output_x : torch.Tensor | list[torch.Tensor]
            output tensor or tensor list
        output_a : torch.Tensor | list[torch.Tensor]
            output a tensor or tensor list
        output_b : torch.Tensor | list[torch.Tensor]
            output b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            lower bound of the truncated interval
        u : torch.Tensor | list[torch.Tensor]
            upper bound of the truncated interval
        """
        output_x = self.forward(x)
        output_a = self.forward(a)
        output_b = self.forward(b)
        return output_x, output_a, output_b, l, u


class GaussianFilter:
    def __init__(self, kernel_size, sigma):
        """Apply a Gaussian filter to the input tensor.
        
        Parameters
        ----------
        kernel_size : int
            size of the Gaussian kernel
        sigma : float
            standard deviation of the Gaussian kernel        
        """
        super().__init__()
        self.kernel_size = kernel_size
        self.sigma = sigma
        self.padding = kernel_size // 2

    def _create_gaussian_kernel(self, dim):
        if dim == 3:  # 1D input
            x = torch.arange(-self.kernel_size // 2 + 1, self.kernel_size // 2 + 1)
            kernel = torch.exp(-(x**2) / (2 * self.sigma**2))
            kernel = kernel / kernel.sum()
            return kernel.unsqueeze(0).unsqueeze(0).double()
        else:  # 2D input
            x = torch.arange(-self.kernel_size // 2 + 1, self.kernel_size // 2 + 1)
            x = x.unsqueeze(0).repeat(self.kernel_size, 1)
            y = x.transpose(0, 1)
            kernel = torch.exp(-(x**2 + y**2) / (2 * self.sigma**2))
            kernel = kernel / kernel.sum()
            return kernel.unsqueeze(0).unsqueeze(0).double()

    def forward(self, x):
        """Compute the convolution of the input tensor with a Gaussian kernel.

        Parameters
        ----------
        x : torch.Tensor
            input tensor

        Returns
        -------
        output : torch.Tensor
            output tensor
        """
        kernel = self._create_gaussian_kernel(x.dim())
        if x.dim() == 3:  # 1D input
            return F.conv1d(
                x,
                kernel.expand(x.shape[1], -1, -1),
                padding=self.padding,
                groups=x.shape[1],
            )
        else:  # 2D input
            return F.conv2d(
                x,
                kernel.expand(x.shape[1], -1, -1, -1),
                padding=self.padding,
                groups=x.shape[1],
            )

    def forward_si(self, x, a, b, l, u, z):
        """
        Parameters
        ----------
        x : torch.Tensor | list[torch.Tensor]
            input tensor or tensor list
        a : torch.Tensor | list[torch.Tensor]
            a tensor or tensor list
        b : torch.Tensor | list[torch.Tensor]
            b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            l tensor or tensor list
        u : torch.Tensor | list[torch.Tensor]
            u tensor or tensor list
        z : float

        Returns
        -------
        output_x : torch.Tensor | list[torch.Tensor]
            output tensor or tensor list
        output_a : torch.Tensor | list[torch.Tensor]
            output a tensor or tensor list
        output_b : torch.Tensor | list[torch.Tensor]
            output b tensor or tensor list
        l : torch.Tensor | list[torch.Tensor]
            lower bound of the truncated interval
        u : torch.Tensor | list[torch.Tensor]
            upper bound of the truncated interval
        """
        output_x = self.forward(x)
        output_a = self.forward(a)
        output_b = self.forward(b)
        return output_x, output_a, output_b, l, u



from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import torch
from onnx import ModelProto
from sicore import (
    SelectiveInferenceNorm,
    SelectiveInferenceChi,
    SelectiveInferenceResult,
)

from . import nn
from .utils import to_numpy
from .hypothesis import PresetHypothesis


@dataclass
class InferenceResult(SelectiveInferenceResult):
    """A class extending SelectiveInferenceResult with ROI and output information.

    Attributes
    ----------
    roi : np.ndarray
        Region of interest for the inference.
    output : np.ndarray
        Output of the inference process.
    score_map : np.ndarray
        Score map obtained from the inference process.
    non_roi : np.ndarray
        Region outside of the ROI.
    stat : float
        Test statistic value.
    p_value : float
        Selective p-value.
    inf_p : float
        Lower bound of selective p-value.
    sup_p : float
        Upper bound of selective p-value.
    searched_intervals : list[list[float]]
        Intervals where the search was performed.
    truncated_intervals : list[list[float]]
        Intervals where the selected model is obtained.
    search_count : int
        Number of times the search was performed.
    detect_count : int
        Number of times the selected model was obtained.
    null_rv : rv_continuous
        Null distribution of the unconditional test statistic.
    alternative : Literal["two-sided", "less", "greater"]
        Type of the alternative hypothesis.
    """

    output: np.ndarray
    score_map: np.ndarray
    roi: np.ndarray
    non_roi: np.ndarray | None = None


class SIModel(ABC):
    def __init__(self):
        self.si_calculator: SelectiveInferenceNorm | SelectiveInferenceChi = None

    @abstractmethod
    def construct_hypothesis(
        self,
        input: torch.Tensor | np.ndarray | list | tuple,
        var: int | float | np.ndarray | torch.Tensor,
        **kwargs,
    ):
        """Abstruct method for construct hypothesis from the observed output of NN.

        Parameters
        ----------
        input : torch.Tensor | np.ndarray | list | tuple
            Input of NN
        var : int | float | np.ndarray | torch.Tensor
            Covariance matrix of input
            Treated as the diagonal of the covariance matrix, representing independent variances for each dimension.

        Raises
        ------
        NoHypothesisError
            If the hypothesis is not obtained from observartion, please raise this error
        """
        pass

    @abstractmethod
    def algorithm(
        self, a: torch.Tensor, b: torch.Tensor, z: float, **kwargs
    ) -> tuple[object, tuple[float, float]]:
        """
        Parameters
        ----------
        a : torch.Tensor
            A vector of nuisance parameter
        b : torch.Tensor
            A vector of the direction of test statistic
        z : float
            A test statistic

        Returns
        -------
        tuple[object, tuple[float,float]]
            First Elements is outputs obtained in the value of z. Second Element is a obtained truncated interval
        """
        pass

    @abstractmethod
    def model_selector(
        self,
        roi_vector: torch.Tensor | np.ndarray | list | tuple | int | float,
        **kwargs,
    ) -> bool:
        """Abstruct method for compare whether same model are obtained from output and observed output(self.output)

        Parameters
        ----------
        roi_vector : Any
            roi obtained from the output of NN

        Returns
        -------
        bool
            If same models are obtained from output and observed output(self.output), Return value should be true. If not, return value should be false.
        """
        pass

    def forward(self, input):
        return self.si_model.forward(input)

    def inference(self, input, var, **kwargs) -> SelectiveInferenceResult:
        self.construct_hypothesis(input, var)
        result = self.si_calculator.inference(
            algorithm=self.algorithm,
            model_selector=self.model_selector,
            **kwargs,
        )
        return result


class PresetSIModel(SIModel):
    def __init__(
        self,
        model,
        hypothesis,
        seed: int = None,
        memoization: bool = True,
        **kwargs,
    ):
        self.si_model = nn.NN(model=model, seed=seed, memoization=memoization)
        self.hypothesis = hypothesis
        self.si_calculator = None
        self.output = None
        self.score_map = None
        self.roi = None

    def construct_hypothesis(self, input, var, **kwargs):
        self.hypothesis.construct_hypothesis(self.si_model, input, var, **kwargs)
        self.si_calculator = self.hypothesis.si_calculator
        self.output = self.hypothesis.output
        self.score_map = self.hypothesis.score_map
        self.roi = self.hypothesis.roi
        if hasattr(self.hypothesis, "non_roi"):
            self.non_roi = self.hypothesis.non_roi
        else:
            self.non_roi = None

    def algorithm(self, a, b, z, **kwargs):
        return self.hypothesis.algorithm(self.si_model, a, b, z, **kwargs)

    def model_selector(self, roi, **kwargs):
        return self.hypothesis.model_selector(roi, **kwargs)

    def inference(
        self,
        input: torch.Tensor
        | np.ndarray
        | list[torch.Tensor | np.ndarray]
        | tuple[torch.Tensor | np.ndarray, ...],
        var: float | np.ndarray | torch.Tensor,
        **kwargs,
    ) -> InferenceResult:
        """
        Parameters
        ----------
        input : torch.Tensor | np.ndarray | list[torch.Tensor | np.ndarray] | tuple[torch.Tensor | np.ndarray, ...]
            Input of NN
        var : float | np.ndarray | torch.Tensor
            Covariance matrix of the noise of input
            Treated as the diagonal of the covariance matrix, representing independent variances for each dimension.
        **kwargs : Any

        Returns
        -------
        InferenceResult
            Result of Selective Inference
        """
        self.construct_hypothesis(input, var)
        result = self.si_calculator.inference(
            algorithm=self.algorithm,
            model_selector=self.model_selector,
            **kwargs,
        )
        result.output = to_numpy(self.output)
        result.score_map = to_numpy(self.score_map)
        result.roi = to_numpy(self.roi)
        result.non_roi = to_numpy(self.non_roi)
        return result


def load(
    model: ModelProto,
    hypothesis: PresetHypothesis,
    seed: float = None,
    memoization: bool = True,
) -> SIModel:
    """Load onnx model and hypothesis setting to SIModel

    Parameters
    ----------
    model : onnx.ModelProto
        The onnx model instance.
    hypothesis : Hypothesis | PresetHypothesis
        The hypothesis setting.
        You can choose an instance of the class "Hypothesis" or class "PresetHypothesis" for preset hypothesis setting.
    seed : float, optional
        The seed of random number generator.
        If the onnx model contains RandomNormalLike layers, the seed is used to generate the same random numbers.
        Default to None.
    memoization : bool, optional
        Whether to use memoization.
        If True, the memoization is enabled

    Returns
    -------
    si_model : SIModel
        The instance of SIModel or SI

    Raises
    ------
    ValueError
        If hypothesis is not an instance of Hypothesis or PresetHypothesis
    """

    if isinstance(hypothesis, PresetHypothesis):
        si_model = PresetSIModel(model, hypothesis, seed, memoization)
    else:
        raise ValueError("hypothesis should be an instance of PresetHypothesis")
    return si_model
